__title__ = 'audible'
__description__ = 'A(Sync) Interface for internal Audible API written in pure Python.'
__url__ = 'https://github.com/mkb79/audible'
__version__ = '0.2.1'
__author__ = 'mkb79'
__author_email__ = 'mkb79@hackitall.de'
__license__ = 'AGPL'
__status__ = 'Development'


__all__ = [
    "__title__", "__description__", "__url__", "__version__",
    "__author__", "__author_email__", "__license__", "__status__"
]
